#include <xsTypes.h>
#include <XSstreams.h>
#include <XSContainer.h>
#include <XSFit/Fit/Fit.h>
#include <XSFit/Fit/StatMethod.h>
#include <XSModel/Model/Model.h>
#include <XSModel/Parameter/ModParam.h>
#include <XSUser/Global/Global.h>
#include <XSUser/Global/XSGlobal.h>
#include <XSUtil/Parse/XSparse.h>
#include <XSUtil/Utils/XSutility.h>
#include <string>
#include <sstream>

int
XSGlobal::xsBayes(ClientData cdata,Tcl_Interp* tclInterp,int objc, Tcl_Obj* CONST  objv[])
{
   enum SUBCOMS {OFF=1, ON, CONS};
   typedef std::map<string, ModParam::PriorType> PriorTypeMap;
   const char* cmd = static_cast<const char*>(cdata);
   static std::map<string, int> subComs;
   static PriorTypeMap priorTypes;
   static string defModName("");
   static size_t defParNum = 1;
   static string defPriorType("CONS");
   static Real defHyperParam = 1.0;
   
   if (!subComs.size())
   {
      subComs["off"] = OFF;
      subComs["on"] = ON;
      subComs["cons"] = CONS;
   }
   if (!priorTypes.size())
   {
      priorTypes["CONS"] = ModParam::CONS;
      priorTypes["EXP"] = ModParam::EXP;
   }
   if (objc == 1)
   {
      XSGlobal::printDocs(cmd,"?");
      return globalData->autoSave(TCL_OK);
   }
   
   try
   {
      IntegerArray iParams(0);
      StringArray xsParams(0);
      std::vector<string> args(objc-1,"");
      for (int i=1; i<objc; ++i) args[i-1] = string(Tcl_GetString(objv[i]));
      XSparse::collectParams(args, iParams, xsParams);

      string firstArg;
      if (iParams[0] == 0)  firstArg = xsParams[0];
      if (firstArg == "?")
      {
         XSGlobal::printDocs(cmd,"?");
         return globalData->autoSave(TCL_OK);
      }

      // First, need to determine if firstArg is a subcommand or a 
      // param specifier.  If firstArg is left blank, assume
      // user is requesting default param specifier.  
      size_t parNum = string::npos;
      string modName;
      bool isParam = false;
      if (!firstArg.length())
      {
         parNum = defParNum;
         modName = defModName;
         isParam = true;
      }
      else
      {
         if (XSparse::stringIntPair(firstArg, modName, parNum)) isParam = true;
            // If it failed stringIntPair test, subCom is either
            // 1. 'ssss' which may be a number, or 2. 'ssss:tttt'
            // where 'tttt' is not a number.
         else if (modName.length())
         {
            // 2nd case above, must be erroneous input.
            std::ostringstream msg;
            msg << firstArg << " is not a valid subcommand or parameter specifier.";
            throw YellowAlert(msg.str());
         }
         else if (parNum == string::npos)
         {
            // 'ssss' not a number
            isParam = false;
         }
         else
            isParam = true;     
      }
      
      const std::map<int,ModParam*>& varPars = XSContainer::fit->variableParameters();
      std::map<int,ModParam*>::const_iterator itVarPars = varPars.begin();
      std::map<int,ModParam*>::const_iterator itVarParsEnd = varPars.end();
      if (isParam)
      {
         // 1st arg is [<modName>:]parNum
         // Verify all args before making any changes.
         string priorType = defPriorType;
         Real hyperParam = defHyperParam;
         ModParam* modPar = 0;
         // verify fit model parameter
         string testName = modName.length() ? modName : Model::DEFAULT();
         itVarPars = varPars.begin();
         while (itVarPars != itVarParsEnd && !modPar)
         {
            ModParam* testPar = itVarPars->second;
            if (testName == testPar->modelName() && parNum == testPar->index())
            {
               modPar = testPar;
            }
            ++itVarPars;
         }
         if (!modPar)
         {
            std::ostringstream msg;
            msg << "Cannot find fit parameter specified by <modName:>number "
               << firstArg;
            throw YellowAlert(msg.str());
         }
         
         // verify additional priorType and hyperParam args.
         for (size_t i=0; i<iParams.size(); ++i)
         {
            size_t iPar = iParams[i];
            if (iPar == 1)
            {
               priorType = xsParams[i];
               string upPriorType = XSutility::upperCase(priorType);
               PriorTypeMap::const_iterator itTypes = priorTypes.lower_bound(upPriorType);
               PriorTypeMap::const_iterator itTypesEnd = priorTypes.end();
               if (itTypes == itTypesEnd || 
                   itTypes->first.find(upPriorType) != 0)
               {
                  tcerr << "  " << priorType << " is an unrecognized Prior type.\n"
                     << "  Choose from the following Prior options (currently \""
                     << defPriorType << "\"):" << std::endl;
                  itTypes = priorTypes.begin();
                  while (itTypes != itTypesEnd)
                  {
                     tcerr << "  " << itTypes->first;
                     ++itTypes;
                  }
                  tcerr << std::endl;
                  return globalData->autoSave(TCL_ERROR);
               }
            }
            else if (iPar == 2)
            {
               std::istringstream iss(xsParams[i]);
               if (!(iss >> hyperParam) || !iss.eof())
               {
                  string msg("Must enter a valid floating-point number for hyper-param value.");
                  throw YellowAlert(msg);
               }
            }            
         }
         
         priorType = XSutility::upperCase(priorType);
         if (priorTypes[priorType] == ModParam::EXP &&
              hyperParam <= 0.0)
         {
            string msg("EXP Prior type requires a finite positive hyper param value.");
            throw YellowAlert(msg);
         }
         
         // If it made it here, all args should be valid.
         modPar->priorType(priorTypes[priorType]);
         modPar->hyperParam(hyperParam);
         defPriorType = priorType;
         defModName = modName;
         defParNum = parNum;
         defHyperParam = hyperParam;
      }
      else
      {
         // By this point, assume 1st arg is a sub-command, not
         // a model parameter. 
         firstArg = XSutility::lowerCase(firstArg);
         int choice = -1;
         std::map<string,int>::const_iterator itSubCom(subComs.lower_bound(firstArg));
         std::map<string,int>::const_iterator itSubComEnd(subComs.end());
         if (itSubCom != itSubComEnd && itSubCom->first.find(firstArg) == 0)
                   choice = itSubCom->second;
         switch (choice)
         {
            using XSContainer::fit;
            case OFF:
               fit->statMethod()->setBayesOn(false);
               break;
            case ON:
               fit->statMethod()->setBayesOn(true);
               break;
            case CONS:
               {
                  itVarPars = varPars.begin();
                  while (itVarPars != itVarParsEnd)
                  {
                     itVarPars->second->priorType(ModParam::CONS);
                     ++itVarPars;
                  }
                  fit->statMethod()->setBayesOn(true);
               }
               break;
            default:
               tcout << "   Subcommand " << firstArg << " does not exist" << std::endl;
               XSGlobal::printDocs(cmd,"?");
               break;
         }
      }
   }
   catch (YellowAlert&)
   {
      return globalData->autoSave(TCL_ERROR);
   }
   return globalData->autoSave(TCL_OK);
}
